package RPM::Source::Convert::Plugins::Python;
use strict;
use warnings;

use RPM::Source::Tools::Dependency::Filter;
use RPM::Source::Transformation::Factory;
use RPM::Source::Transformation::OptHandlerWrapper;
use RPM::Source::Transformation::Factory::DependencyFilter;
use RPM::Source::Convert::Plugins::SRPMReset;

our @ISA=qw/RPM::Source::Transformation::Factory/;
push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our $verbose=0;

# see also https://docs.fedoraproject.org/en-US/packaging-guidelines/Python/

&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'Python PreDistromap',
    -description => 'preconvert python macro in Python package names before distromap',
    -class => 'RPM::Source::Tools::Dependency::Filter::PythonPreDistromap',
    -tags => SPEC_TAG_SET_REQUIRES | SPEC_TAG_SET_CONFLICTS,
    -priority => FILTER_PRIORITY_PREDISTROMAP,
);

&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'Python PostDistromap',
    -description => 'convert generic Python name conventions to alt name conventions',
    -class => 'RPM::Source::Tools::Dependency::Filter::PythonPostDistromap',
    -tags => SPEC_TAG_SET_REQUIRES | SPEC_TAG_SET_CONFLICTS,
    -priority => FILTER_PRIORITY_POSTDISTROMAP,
);


sub transformations {
    return
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_SRPM_REINIT,
	    -name => 'generic_python_related_reinit',
	    -opt_handler_sub=>\&opt_handler_python_reinit),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_PRECONVERT,
	    -name => 'generic_python_related_preconvert',
	    -opt_handler_sub=>\&opt_handler_python_preconvert),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_CONVERT,
	    -name => 'generic_python_related_convert',
	    -opt_handler_sub=>\&opt_handler_python_convert),
	;
}

# https://www.python.org/dev/peps/pep-0503/#normalized-names
sub __pyname_to_canonical {
    my $pyname=lc($_[0]);
    $pyname=~s,[-_.]+,-,g;
    return $pyname;
}

sub __pyname_to_3dist {
    return 'python3dist('.&__pyname_to_canonical($_[0]).')';
}

sub __pynames_to_3dist {
    my $instring=$_[0];
    $instring=~s,^\s*,,;
    $instring=~s,\s*$,,;
    return join(' ',map {&__pyname_to_3dist($_)} split(/\s+/,$instring));
}

# Fedora macro to replace overly complicated references to PyPI source files.
# Expands to the pythonhosted URL for a package
# Accepts zero to three arguments:
# 1:  The PyPI project name, defaulting to %%srcname if it is defined, then
#     %%pypi_name if it is defined, then just %%name.
# 2:  The PYPI version, defaulting to %%version with tildes stripped.
# 3:  The file extension, defaulting to "tar.gz".  (A period will be added
#     automatically.)
# Requires %%__pypi_url and %%__pypi_default_extension to be defined.
#%__pypi_url https://files.pythonhosted.org/packages/source/
#%__pypi_default_extension tar.gz
sub __pypi_source {
    my ($macros,$argstring) = @_;
    my ($src,$ver,$ext);
    if (defined $argstring) {
	$argstring=~s/\}$//;
	($src,$ver,$ext)=split(/\s+/,$argstring);
    }
    my $url='%__pypi_url';
    $url='https://files.pythonhosted.org/packages/source/' unless $macros->get_raw('__pypi_url');
    # If no first argument, try %srcname, then %pypi_name, then %name
    unless (defined($src)) {
	if ($macros->get_raw('srcname')) {
	    $src = '%{srcname}';
	} elsif ($macros->get_raw('pypi_name')) {
	    $src = '%{pypi_name}';
	} else {
	    $src = '%{name}';
	}
    }
    # If no second argument, use %version
    unless (defined($ver)) {
	# TODO: if ~ will be allowed...
        #$ver = rpm.expand('%version'):gsub('~', '')
        $ver = '%{version}';
    }

    # If no third argument, use the preset default extension
    unless (defined($ext)) {
	$ext='%__pypi_default_extension';
	$ext='tar.gz' unless $macros->get_raw('__pypi_default_extension');
    }
    my $first=substr($macros->macro_subst($src), 0, 1);
    print STDERR 'DEBUG: url=',$url,' first=',$first,' src=',$src,' ver=',$ver,' ext=',$ext,"\n" if $verbose > 3;
    return $url . $first . '/' . $src . '/' . $src . '-' . $ver . '.' . $ext;
}

# TODO: fc specific? will it clash with mageia?
sub opt_handler_python_reinit {
    my (undef,$spec,undef,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    # for proper conversion of subpackage names
    $macros->set_raw('python3_pkgversion',3);

    #$spec->main_section->map_body(sub{
    #  s,\%\{?pypi_source\s+(\S+)\}?(?=\s*$),https://pypi.org/project/$1/$1-%{version}.tar.gz,;
    #                    });

    $spec->main_section->map_body(
	sub {
	    # fc specific
	    if (s!\%\{?pypi_source(?:\s+(\S.*))?(?:\}|$)!&__pypi_source($macros,$1)!ge) {
		# Source0: %pypi_source
		# no need
		#$opt->{-srpm_reset_speclist_after_reinit}=1;
		# set sourceXX in macro
		$opt->{-srpm_reset_sources_after_reinit}=1;
	    }
	    # ---
	} );
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub opt_handler_python_preconvert {
    my (undef,$spec,undef,$opt,$resourceobj)=@_;
    #my $macros=$spec->macros();
    foreach my $sec ($spec->get_sections()) {
	my $t = $sec->get_type();
	next if $t eq 'changelog';
	if ($t eq 'package') {
	    $sec->map_body(sub {
		# fc specific
		s,\%\{py3_dist\s+([^\}]+)\s*\},&__pynames_to_3dist($1),ge;
		# ---
			   } );

	}
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}

sub opt_handler_python_convert {
    my (undef,$spec,undef,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    my $mainsec=$spec->main_section;
    my $buildreq=$resourceobj->buildreq;

    my ($flag_macro_python2,$flag_macro_python3);
    foreach my $sec ($spec->get_sections()) {
	my $t = $sec->get_type();
	next if $t eq 'changelog';
	if ($t eq 'package') {
	    $sec->map_body(sub {
		s,\%\{py3_dist\s(\S+)\s*\},&__pyname_to_3dist($1),ge;
		s!\%\{?pypi_source(?:\s+(\S.*))?(?:\}|$)!&__pypi_source($macros,$1)!ge;
			   });
	} elsif ($t eq 'build') {
	    $sec->map_body(sub {
		# beware of %{?with_python2:%py2_build}
		# warning: setuptools is a hack, remove me when fixed in SourceAnalyser
		# python2 is going to be dropped ..
		#$buildreq->add_buildreq('python-devel','python-module-setuptools')
		#    if s/(?<!%)\%\{py2?_build\}/%python_build/g or
		#       s/(?<!%)\%py2?_build\b/%python_build/g;
		s/(?<!%)\%\{py2?_build\}/%python_build/g;
		s/(?<!%)\%py2?_build\b/%python_build/g;

		$buildreq->add_buildreq('python3-devel','python3-module-setuptools')
		    if s/(?<!%)\%\{py3_build\}/%python3_build/g or
		       s/(?<!%)\%py3_build\b/%python3_build/g;
			   });
	} elsif ($t eq 'install') {
	    $sec->map_body(sub {
		# beware of %{?with_python2:%py2_install}
		s/(?<!%)\%py2?_install\b/%python_install/g;
		s/(?<!%)\%py3_install\b/%python3_install/g;
		s/(?<!%)\%\{py2?_install\}/%python_install/g;
		s/(?<!%)\%\{py3_install\}/%python3_install/g;

		# HACK ... not exactly a direct translation ... and does not work
		# Byte compile the python files
		#%py_byte_compile %{__python} $RPM_BUILD_ROOT%{pkgdir}/xemacs-packages/etc/python-modes
		#$flag_macro_python2=1 if s/^\s*\%py_byte_compile\s+\%\{__python2?\}\s+\$RPM_BUILD_ROOT/%add_python_compile_include /;
			   });

	} elsif ($t eq 'files') {
	    #
	    $sec->map_body(sub {
		# python3 packaging (in Fedora)
		s,\.cpython-\%\{python3_version_nodots\}m(?:-\*)?\.so\b,%{_python3_extension_suffix},;
		# python3 packaging TMP HACK!!! TODO!!!
		s,\.cpython-\%\{python3_version_nodots\}\.,.cpython-*.,;
		s,\.cpython-\%\{python3_version_nodots\}\*,.cpython-3*,;
		s,\%\{python3_version_nodots\}\*,3*,;

		# %pycached()
		if (s,^\s*%pycached\s+,, && m!^(.*/)([^/]+)(?:\.py)\s*$!) {
		    my $pydir=$1;
		    my $pyprefix=$2;
		    $_ .= $pydir.'__pycache__/'.$pyprefix .'.cpython-3*.pyc'."\n";
		    $_ .= $pydir.'__pycache__/'.$pyprefix .'.cpython-3*.opt-?.pyc'."\n";
		}
			   } );

	}

	$sec->map_body(sub {
	    s,(?<!%)\%\{python3\},/usr/bin/python3,g;
	    s,(?<!%)\%python3(\s),/usr/bin/python3$1,g;
	    # py.test-%{python3_version}
	    s/\bpy\.?test-\%\{?(?:__)?python3_version\}?/py.test-3/g;
	    s/(?<!%)\%pytest(\s)/py.test-3$1/g;
	    s/\btrial-\%\{?(?:__)?python3_version\b\}?/trial-3/g;
	} ) if ($t eq 'install' or $t eq 'build' or $t eq 'prep' or $t eq 'check');



	# general hacks here
	$sec->map_body(sub {

	    if (/(?<!%)\%\{?python2?_site(?:lib|arch)\}?(?!dir)/) {
		s,\%\{python2?_sitelib\},%{python_sitelibdir_noarch},g;
		s,\%python2?_sitelib,%python_sitelibdir_noarch,g;
		s,\%\{python2?_sitearch\},%{python_sitelibdir},g;
		s,\%python2?_sitearch,%python_sitelibdir,g;
		$flag_macro_python2=1;
	    }
	    if (/(?<!%)\%\{?__python2\}?/) {
		s,\%\{__python2\},%{__python},g;
		s,\%__python2(\s),%{__python}$1,g;
		$flag_macro_python2=1;
	    }

	    if (/(?<!%)\%\{?python3_site(?:lib|arch)\}?(?!dir)/) {
		s,(?<!%)\%\{python3_sitelib\},%{python3_sitelibdir_noarch},g;
		s,(?<!%)\%python3_sitelib,%python3_sitelibdir_noarch,g;
		s,(?<!%)\%\{python3_sitearch\},%{python3_sitelibdir},g;
		s,(?<!%)\%python3_sitearch,%python3_sitelibdir,g;
		$flag_macro_python3=1;
	    }
	    if (
		s,(?<!%)\%\{py3dir\},%{_builddir}/python3-%{name}-%{version}-%{release},g or
		s,(?<!%)\%py3dir,%{_builddir}/python3-%{name}-%{version}-%{release},g) {
		$flag_macro_python3=1;
	    }
	    $flag_macro_python3=1 if m'(?<!%)\%\{?__python3\}?';

	    s,(?<!%)\%\{python_version\},%{__python_version},g if not $macros->get_raw('python_version');
	    s,(?<!%)\%\{python2_version\},%{__python_version},g if not $macros->get_raw('python2_version');
	    s,(?<!%)\%\{python3_version\},%{__python3_version},g if not $macros->get_raw('python3_version');

	    s,(?<!%)\%\{python3_pkgversion\},3,g;
		       } );

    }
    if ($flag_macro_python2) {
	$buildreq->add_pre_buildreq('rpm-build-python');
	$buildreq->add_buildreq('python-devel');
    }
    if ($flag_macro_python3) {
	$buildreq->add_pre_buildreq('rpm-build-python3');
	$buildreq->add_buildreq('python3-devel');
    }
    return RPM::Source::TransformContainer::TRANSFORMATION_OK;
}


package RPM::Source::Tools::Dependency::Filter::PythonPreDistromap;
use strict;
use warnings;
our @ISA=q/RPM::Source::Tools::Dependency::Filter/;

sub apply {
    my ($self, $tagobj, $dep) = @_;
    my $macros=$self->{-spec}->macros;
    my $pkgname=$dep->{-name};
    # noticed in fedora / epel
    $pkgname=~s,\%\{python3_pkgversion\},3,g;
    $dep->{-name}=$pkgname;
    return $dep;
}

package RPM::Source::Tools::Dependency::Filter::PythonPostDistromap;
use strict;
use warnings;
our @ISA=q/RPM::Source::Tools::Dependency::Filter/;

sub apply {
    my ($self, $tagobj, $dep) = @_;
    my $pkgname=$dep->{-name};
    if ($pkgname=~/^python[23]?-(?:base$|contextkit$|dev$|devel$|test$|tools|module-)/) {
	# skip below branch
    } elsif ($pkgname=~s/^python2-/python-module-/) {
	$dep->{-name}=$pkgname;
    } elsif ($pkgname=~s/^python3-/python3-module-/) {
	$dep->{-name}=$pkgname;
    }
    return $dep;
}

1;
