package RPM::Source::Convert::Tools::Fonts;

use strict;
use warnings;
use Carp;
use Cwd;
use File::Basename;
use Data::Dumper;
use RPM::Source::Convert::Tools::LuaFedoraCommon;

our $verbose=0;
# fedora-style=ends_with_fonts
our $fontname_ends_with_fonts=0;

=head
%global foundry           GFS
%global fontlicense       OFL
%global fontlicenses      OFL.txt
%global fontdocs          *.txt
%global fontdocsex        %{fontlicenses}

%global fontfamily        Eustace
%global fontsummary       GFS Eustace, a majuscule Greek font family
%global fonts             *.otf
%global fontconfngs       %{SOURCE10}
%global fontdescription   \
As it is known, the Greek alphabet ...

%global archivename GFS_Eustace


=cut

sub __get_old_name {
    my ($spec,$macros,$resourceobj)=@_;
    confess "Oops! __get_old_name: got UnDEFFF" unless $resourceobj;
    my $name=$resourceobj->datapath_get_name();
    return $name if defined($name) and $name ne '';
    if ($macros->get_raw('name')) {
	$name=$macros->macro_subst('%name');
    } else {
	($name,)=$spec->_get_nevr();
	if (!defined($name)) {
	    die 'ERROR: FedoraFontPkg2ALT: main package name not found. Please, define %name from command line.';
	}
    }
    # due to fontforge, the name is not set in prepare_resource; set it forcefully
    $resourceobj->datapath_set_name($name);
    print STDERR "Tools::Fonts: datapath_set_name to ",$name,"\n" if $verbose;
    return $name;
}

# Return a normalized name
sub __norm {
    my $r = shift;
    $r=~s,[[:punct:]\s]+,-,g;
    $r=~s/^-//;
    $r=~s/-$//;
    return lc($r);
}

# loop over suffixlist and return name minus the first suffix that matches
# - is used as suffix separator
# name should have passed through norm at one point in the past
sub __dropsuffix {
    my ($name,$suffixlist)=@_;
    my $r = $name;
    for my $s (@$suffixlist) {
	my $norm_s=&__norm($s);
	last if $r =~ s/-$norm_s$//;
    }
    return $r;
}



# Compute a font family name that can be used in packaging, lowercasing, using
# - as separator, and applying  “WPF font selection model” whitepaper
# simplifications
sub __rpmname {
    my ($name)=@_;
    my $r = &__norm($name);
    # Normal & co
    $r = &__dropsuffix($r,["normal","book","regular","upright"]);
    # Slant
    $r = &__dropsuffix($r,["italic","ita","ital","cursive","kursiv",
                    "oblique","inclined","backslanted","backslant","slanted"]);
    # Width / Stretch
    $r = &__dropsuffix($r,["ultracondensed","extra-compressed","ext-compressed","ultra-compressed","ultra-condensed",
                    "extracondensed","compressed","extra-condensed","ext-condensed","extra-cond",
                    "semicondensed","narrow","semi-condens",
                    "semiexpanded","wide","semi-expanded","semi-extended",
                    "extraexpanded","extra-expanded","ext-expanded","extra-extended","ext-extended",
                    "ultraexpanded","ultra-expanded","ultra-extended",
                    "condensed","cond",
                    "expanded","extended"]);
    # Weight (no abbreviated suffix handling, too dangerous)
    $r = &__dropsuffix($r,["thin","extra-thin","ext-thin","ultra-thin",
                    "extralight","extra-light","ext-light","ultra-light",
                    "demibold","semi-bold","demi-bold",
                    "extrabold","extra-bold","ext-bold","ultra-bold",
                    "extrablack","extra-black","ext-black","ultra-black",
                    "bold","thin","light","medium",
                    "black","heavy","nord",
                    "demi","ultra"]);

    my %tokens = map {$_=>1} ("font","fonts");
    my @ts = split(/[%-]+/,$r);
    $r = '';
    for my $t (@ts) {
	if (not $tokens{$t}) {
	    $r .= "-" . $t;
	    $tokens{$t} = 1;
	}
    }
    $r =~ s,^-,,;
    return $r.'-fonts';
}

# calculate fontpkgname
sub __get_fontpkgname {
    my ($spec, $macros, $resourceobj, $suffix)=@_;
    confess 'Oops! __get_fontpkgname: got undef suffix' unless defined($suffix);
    my $fontpkgname;
    my $raw=$macros->get_raw('fontpkgname'.$suffix);
    if (!$suffix && !defined($raw)) {
	if ($suffix eq '') {
	    $raw=$macros->get_raw('fontpkgname0');
	} else {
	    $raw=$macros->get_raw('fontpkgname');
	}
    }
    $fontpkgname=$macros->macro_subst($raw) if defined($raw);
    unless ($fontpkgname) {
	my $foundry=&get_current($spec,$macros,'foundry',$suffix);
	my $fontfamily=&get_current($spec,$macros,'fontfamily',$suffix);
	my $basename=($foundry ? $foundry.' '.$fontfamily : $fontfamily);
	$fontpkgname=&__rpmname($basename) if $basename;
	#warn 'fontpkgname=',$fontpkgname,' suffix=',$suffix,"\n";
	$fontpkgname//=&__get_old_name($spec,$macros,$resourceobj) unless $suffix;
    } else {
	#warn 'SSSS: fontpkgname=',$fontpkgname,' suffix=',$suffix,"\n";
    }
    return $fontpkgname;
}

sub __get_fontfilelist {
    my ($spec, $macros, $resourceobj, $suffix)=@_;
    return &get_current($spec,$macros,'fontfilelist',$suffix,&__get_fontpkgname($spec,$macros,$resourceobj,$suffix).$suffix.'.list');
}

sub _fontfiles {
    my ($spec,$macros,$resourceobj,$suffix)=@_;
    my $pkgname=&__get_fontpkgname($spec,$macros,$resourceobj, $suffix);
    my $fontstem=&__get_alt_fontstem($spec,$macros,$resourceobj,$suffix);
    $macros->set_raw('_fontdir','%_fontsdir/*/'.$fontstem);
    return '%files -n '.$pkgname.' -f '.&__get_fontfilelist($spec,$macros,$resourceobj, $suffix)."\n";
}

# converted from Lua code used by macros.fonts and derivatives

sub get_current {
    my ($spec,$macros,$name,$suffix,$default)=@_;
    $suffix //= '';
    my $ismain = ($suffix eq '' or $suffix eq '0');
    my $def_rawval=$macros->get_raw($name);
    my $raw_val;
    if ($ismain) {
	$raw_val=$def_rawval;
	$raw_val//=$macros->get_raw($name.'0');
    } else {
	$raw_val=$macros->get_raw($name.$suffix);
    }
    # for example, %fontorg: bare macros first, then default
    $raw_val//=$def_rawval if defined($def_rawval);
    $raw_val//=$default if defined($default);
    return $raw_val unless defined($raw_val);
    if ($raw_val=~/^\%\{?SOURCE(\d*)\}?$/i) {
	return $raw_val;
	#my $n=$1;
	#$n=0 if $n eq '';
	#return $macros->macro_subst($spec->raw_sourcelist_ref->[$n]);
    }
    return $macros->macro_subst($raw_val);
}

sub fc_macros_listfiles {
    my ($listfiles_include, $listfiles_exclude,$noglob)=@_;
    #confess "Oops: fc_macros_listfiles: undefined listfiles_include" if ! defined $listfiles_include;
    return unless defined $listfiles_include;
    carp "Oops: fc_macros_listfiles: newline in include" if $listfiles_include =~ /\n/;
    carp "Oops: fc_macros_listfiles: newline in exclude" if defined($listfiles_exclude) and $listfiles_exclude =~ /\n/;
    my %_exclude;
    %_exclude=map{$_=>1} glob($listfiles_exclude) if defined $listfiles_exclude;
    my @retval;
    @retval=grep {(/^\%\{?SOURCE(\d*)\}?$/i or -e $_ or ($noglob and not /\*/)) and not $_exclude{$_}} glob($listfiles_include);
    return @retval;
}


sub closetag {
    my ($oldtag, $newtag)=@_;
    return '' if ! defined $oldtag;
    my $output = ']]></' . $oldtag . '>';
    if ($oldtag eq 'li' and $newtag ne $oldtag) {
	$output .= '</ul>';
    }
    return $output;
}

sub opentag {
    my ($oldtag, $newtag) =@_;
    return '' if !defined $newtag;
    my $output = '<' . $newtag . '><![CDATA[';
    if ($newtag eq 'li' and $newtag ne $oldtag) {
	$output = '<ul>' . $output;
    }
    return $output;
}

sub switchtag {
    my ($oldtag, $newtag) = @_;
    return &closetag($oldtag, $newtag) . &opentag($oldtag, $newtag);
}

# Reformat some text into something that can be included in an AppStream
# XML description
sub txt2xml {
    my ($text)=@_;
    $text=&RPM::Source::Convert::Tools::LuaFedoraCommon::wordwrap($text);
    my $output = '';
    my $oldadvance = '';
    my ($oldtag,$newtag,$advance);
    $text =~ s,^\n*,,s;
    $text =~ s,\n*$,,s;
    $text .="\n";
    while ($text =~ s,^([^\n]*\n),,) {
	my $line = $1;
	my $change = 1;
	if ($line=~ m,^(\s*– ),) {
	    $advance = $1;
	    $newtag = "li";
	} elsif  ($line=~ m,^(\s*),,) {
	    $advance = $1;
	    $newtag = '';
	    if ($line eq "\n") {
		$newtag = undef;
	    } elsif ($advance eq $oldadvance) {
		$newtag = "p";
	    } else {
		$change = 0;
	    }
	}
	my $result = $line;
	if ($change) {
	    my $switchtag=&switchtag($oldtag,$newtag);
	    $result=~s,^$advance,$switchtag,ge;
	    $oldtag = $newtag;
	    $oldadvance = $advance=~s,– ,  ,r;
	} else {
	    $result =~ s,^$advance, ,;
	}
	$result =~ s,\n$,,;
	$output .=  $result;
    }
    $output .= &closetag($oldtag, undef);
    return $output;
}

sub __fontbuild {
    my ($spec,$macros,$resourceobj,$suffix,$setupdir)=@_;
    my $curdir = getcwd;
    chdir($setupdir) || die "internal error: failed to chdir to",$setupdir;
    my $listfiles_include=&get_current($spec,$macros,'fontappstreams',$suffix);
    my $listfiles_exclude=&get_current($spec,$macros,'fontappstreamsex',$suffix);
    my @fontappstreams;
    @fontappstreams=&fc_macros_listfiles($listfiles_include,$listfiles_exclude) if $listfiles_include;
    $listfiles_include=&get_current($spec,$macros,'fonts',$suffix);
    $listfiles_exclude=&get_current($spec,$macros,'fontsex',$suffix);
    # noglob=>1 because fonts can be built in %build sec
    my @fonts=&fc_macros_listfiles($listfiles_include,$listfiles_exclude,1);
    chdir($curdir);
    my $output='# fontbuild '.$suffix."\n";
    if (0==@fontappstreams) {
	$output.=q!fontnames=$(
  for font in !.&__shellquote(@fonts).q!; do
    fc-scan "${font}" -f "    <font>%%{fullname[0]}</font>\\n"
  done | sort -u
)
if [[ -n "${fontnames}" ]] ; then
  fontnames=$'\\n'"  <provides>"$'\\n'"${fontnames}"$'\\n'"  </provides>"
fi
fontlangs=$(
  for font in !.&__shellquote(@fonts).q!; do
    fc-scan "${font}" -f "%%{[]lang{    <lang>%%{lang}</lang>\\n}}"
  done | sort -u
)
if [[ -n "${fontlangs}" ]] ; then
  fontlangs=$'\\n'"  <languages>"$'\\n'"${fontlangs}"$'\\n'"  </languages>"
fi
! if @fonts>0;
=head lua
  local basename = rpm.expand("%{?foundry" .. suffix .. ":%{foundry" .. suffix .. "} }%{fontfamily" .. suffix .. "}")
  fedora.safeset("fontpkgname"   .. suffix, rpmname(basename), verbose)
  fedora.safeset("fonthumanname" .. suffix, basename, verbose)
=cut
	my $foundry=&get_current($spec,$macros,'foundry',$suffix);
	my $fontfamily=&get_current($spec,$macros,'fontfamily',$suffix);
	my $basename=($foundry ? $foundry.' '.$fontfamily : $fontfamily);
	my $fontpkgname=&__get_fontpkgname($spec,$macros,$resourceobj,$suffix);
	my $fontlicense=&get_current($spec,$macros,'fontlicense',$suffix);
	my $fonthumanname=&get_current($spec,$macros,'fonthumanname',$suffix,$basename);
	my $fontsummary=&get_current($spec,$macros,'fontsummary',$suffix);
	my $fontdescription=&get_current($spec,$macros,'fontdescription',$suffix);
	my $fontcontact=&get_current($spec,$macros,'fontcontact',$suffix,'devel@lists.altlinux.org');
	my $fontorg=&get_current($spec,$macros,'fontorg',$suffix,'org.altlinux');
	my $mainsec=$spec->main_section;
	my $url=$mainsec->get_tag('URL',RAW=>1);
	$fontlicense//=$mainsec->get_tag('License',RAW=>1);
	$output.=q&
echo "Generating the &.$fontpkgname.q& appstream file"
cat > "&.$fontorg.q&.&.$fontpkgname.q&.metainfo.xml" << EOF_APPSTREAM
<?xml version="1.0" encoding="UTF-8"?>
<!-- SPDX-License-Identifier: MIT -->
<component type="font">
  <id>&.$fontorg.'.'.$fontpkgname.q&</id>
  <metadata_license>MIT</metadata_license>
  <project_license>&.$fontlicense.q&</project_license>
  <name>&.$fonthumanname.q&</name>
  <summary><![CDATA[&.$fontsummary.q&]]></summary>
  <description>
    &.&txt2xml($fontdescription).q&
  </description>
  <updatecontact>&.$fontcontact.q&</updatecontact>
&.($url? '  <url type="homepage">'.$url.'</url>'."\n" : '').
q&  <releases>
    <release version="%{version}-%{release}" date="$(date -d @$SOURCE_DATE_EPOCH -u --rfc-3339=d)"/>
  </releases>${fontnames}${fontlangs}
</component>
EOF_APPSTREAM
&;
    }
    return $output;
}

sub __get_font_ext {
    my ($spec,$font)=@_;
    if ($font=~/^\%\{?SOURCE(\d*)\}?$/i) {
	my $n=$1;
	$n=0 if $n eq '';
	$font=$spec->macros->macro_subst($spec->raw_sourcelist_ref->[$n]);
    }
    my $ext = $font =~ s/^.*\.([\w\d]+)$/$1/r;
    if ($ext eq 'ttc') {
	$ext='ttf';
    } elsif ($ext eq 'TTC') {
	$ext='ttf';
    } elsif ($ext eq 'TTF') {
	$ext='ttf';
    } elsif ($ext eq 'OTF') {
	$ext='otf';
    }
    return $ext;
}

sub __get_alt_fontstem {
    my ($spec,$macros,$resourceobj,$suffix)=@_;
    # do we need it?
    my $oldrpmname=&__get_old_name($spec,$macros,$resourceobj);
    my $fontname=&get_current($spec,$macros,'fontname',$suffix,$oldrpmname);
    $fontname=~s,-fonts$,, unless $fontname_ends_with_fonts;
    return $fontname;
}

sub __fontinstall {
    my ($spec,$macros,$resourceobj,$suffix,$setupdir)=@_;
    my $curdir = getcwd;
    chdir($setupdir) || die "internal error: failed to chdir to",$setupdir;
    my $listfiles_include=&get_current($spec,$macros,'fontappstreams',$suffix);
    my $listfiles_exclude=&get_current($spec,$macros,'fontappstreamsex',$suffix);
    my @fontappstreams;
    @fontappstreams=&fc_macros_listfiles($listfiles_include,$listfiles_exclude) if $listfiles_include;
    $listfiles_include=&get_current($spec,$macros,'fonts',$suffix);
    $listfiles_exclude=&get_current($spec,$macros,'fontsex',$suffix);
    # noglob=>1 because fonts can be built in %build sec
    my @fonts=&fc_macros_listfiles($listfiles_include,$listfiles_exclude,1);
    $listfiles_include=&get_current($spec,$macros,'fontconfngs',$suffix);
    $listfiles_exclude=&get_current($spec,$macros,'fontngsex',$suffix);
    my @fontconfngs;
    @fontconfngs=&fc_macros_listfiles($listfiles_include,$listfiles_exclude) if $listfiles_include;
    print STDERR 'Fonts:%install: incl=',(defined $listfiles_include?$listfiles_include:''),' excl=',(defined $listfiles_exclude ? $listfiles_exclude:''),' @fontconfngs=',join(' ',@fontconfngs),' raw: fontconfngs=',$macros->get_raw('fontconfngs'),"\n" if $verbose>1;
    $listfiles_include=&get_current($spec,$macros,'fontconfs',$suffix);
    $listfiles_exclude=&get_current($spec,$macros,'fontconfsex',$suffix);
    my @fontconfs=&fc_macros_listfiles($listfiles_include,$listfiles_exclude);
    $listfiles_include=&get_current($spec,$macros,'fontdocs',$suffix);
    $listfiles_exclude=&get_current($spec,$macros,'fontdocsex',$suffix);
    my @fontdocs=&fc_macros_listfiles($listfiles_include,$listfiles_exclude);
    $listfiles_include=&get_current($spec,$macros,'fontlicenses',$suffix);
    $listfiles_exclude=&get_current($spec,$macros,'fontlicensesex',$suffix);
    my @fontlicenses=&fc_macros_listfiles($listfiles_include,$listfiles_exclude);
    chdir($curdir);

    my $output='';
    my $foundry=&get_current($spec,$macros,'foundry',$suffix);
    my $fontfamily=&get_current($spec,$macros,'fontfamily',$suffix);
    my $basename=($foundry ? $foundry.$fontfamily : $fontfamily);
    my $fontpkgname=&__get_fontpkgname($spec,$macros,$resourceobj,$suffix);
    my $fontlicense=&get_current($spec,$macros,'fontlicense',$suffix);
    my $fontorg=&get_current($spec,$macros,'fontorg',$suffix,'org.altlinux');
    my $fontfilelist=&__get_fontfilelist($spec,$macros,$resourceobj,$suffix);
    my $fontstem=&__get_alt_fontstem($spec,$macros,$resourceobj,$suffix);
    if (@fonts==0) {
	print STDERR 'ERROR: fontimport: no fonts found (suffix=',$suffix,")\n";
	return '';
    }
    $output.=
'echo Installing '.$fontpkgname.'
echo "" > "'.$fontfilelist.'"
';
    my %fontext2dir=map{$_=>$_} map {&__get_font_ext($spec,$_)} @fonts;
    my @unimplemented=grep {$_ ne 'ttf' and $_ ne 'otf'} keys(%fontext2dir);
    die "Fonts: not implemented: ",join(' ',@unimplemented) if @unimplemented;
    foreach my $ext (keys(%fontext2dir)) {
	my $fontdir=q!%_fontsdir/!.$fontext2dir{$ext}.'/'.$fontstem;
	$output.=q!install -m 0755 -vd %buildroot!.$fontdir.q!/
echo "%%dir !.$fontdir.q!" >> "!.$fontfilelist.q!"
!;
    }
    foreach my $font (@fonts) {
	my $ext = &__get_font_ext($spec,$font);
	#%_fontdir %{_fontsdir}/%{?fontname:%{fontname}}%{?!fontname:%{name}}
	my $fontdir=q!%_fontsdir/!.$fontext2dir{$ext}.'/'.$fontstem.'/';
	$output.=q!install -m 0644 -vp "!.$font.q!" %buildroot!.$fontdir."\n";
	if (substr($font,0,1) eq '%') {
	    $output.=q!echo \\"!.$fontdir.q!$(basename "!.$font.q!")\\" >> '!.$fontfilelist."'\n";
	} else {
	    $output.=q!echo \\"!.$fontdir.basename($font).q!\\" >> '!.$fontfilelist."'\n";
	}
    }
    if (@fontconfs>0 or @fontconfngs>0) {
	$output.="(\n";
	$output.=q!
  IFS= lines=$(
    for fontconfng in !.&__shellquote(@fontconfngs).q!; do
      gen-fontconf -x "${fontconfng}" -w -f !.&__shellquote(@fonts).q!
    done
  )
  while IFS= read -r line; do
    [[ -n $line ]] && newfontconfs+=("$line")
  done <<< ${lines}
! if @fontconfngs>0;
	$output.=q!
  install -m 0755 -vd "%{buildroot}%{_fontconfig_templatedir}" \\
                    "%{buildroot}%{_fontconfig_confdir}"
  for fontconf in !.&__shellquote(@fontconfs).q! "${newfontconfs[@]}"; do
    if [[ -n $fontconf ]] ; then
      install -m 0644 -vp "${fontconf}" "%{buildroot}%{_fontconfig_templatedir}"
      echo \\"%{_fontconfig_templatedir}/$(basename "${fontconf}")\\"                  >> "!.$fontfilelist.q!"
      ln -vsr "%{buildroot}%{_fontconfig_templatedir}/$(basename "${fontconf}")" "%{buildroot}%{_fontconfig_confdir}"
      echo "%%config(noreplace)" \\"%{_fontconfig_confdir}/$(basename "${fontconf}")\\" >> "!.$fontfilelist.q!"
    fi
  done
!;
	$output.=")\n";
    }

    push @fontappstreams, $fontorg.'.'.$fontpkgname.'.metainfo.xml' if 0==@fontappstreams;
    $output.=q!
install -m 0755 -vd "%{buildroot}%{_metainfodir}"
for fontappstream in !.&__shellquote(@fontappstreams).q!; do
  install -m 0644 -vp "${fontappstream}" "%{buildroot}%{_metainfodir}"
  echo \\"%{_metainfodir}/$(basename "${fontappstream}")\\" >> "!.$fontfilelist.q!"
done
! if @fontappstreams>0;
    $output.=q!
for fontdoc in !.&__shellquote(@fontdocs).q!; do
  echo %%doc "'${fontdoc}'" >> "!.$fontfilelist.q!"
done
! if @fontdocs>0;
#    echo %%license "${fontlicense}" >> "!.$fontfilelist.q!"
    $output.=q!
for fontlicense in !.&__shellquote(@fontlicenses).q!; do
  echo %%doc "'${fontlicense}'" >> "!.$fontfilelist.q!"
done
! if @fontlicenses>0;
    return $output;
}

sub __fontcheck {
    my ($spec,$macros,$resourceobj,$suffix,$setupdir)=@_;
    #return '#%fontcheck'."\n";
    my $fontfilelist=&__get_fontfilelist($spec,$macros,$resourceobj,$suffix);
    my $output.=q!# fontcheck !.$suffix.q!
grep -E '^"%{_fontconfig_templatedir}/.+\\.conf"' '!.$fontfilelist.q!' \\
  | xargs -I{} -- sh -c "xmllint --loaddtd --valid     --nonet '%{buildroot}{}' >/dev/null && echo %{buildroot}{}: OK"
grep -E '^"%{_datadir}/metainfo/.+\\.xml"'        '!.$fontfilelist.q!' \\
  | xargs -I{} --        appstream-util validate-relax --nonet '%{buildroot}{}'
!;
    return $output;
}

sub __shellquote {
    return join(' ',map {"'".$_."'"} @_);
}

1;

__END__
