package RPM::Source::Convert::Tools::ForgeMeta;

use strict;
use warnings;
use Carp;
use Data::Dumper;

our $verbose=0;

# https://docs.fedoraproject.org/en-US/packaging-guidelines/SourceURL/#_using_forges_hosted_revision_control
# https://src.fedoraproject.org/rpms/redhat-rpm-config/c/c70110c677904a188d66ffd01a9ab024469ca62d

=head

Name:           naga
Version:        3.0

%global commit  6f1e95ddfdb1aa7719c98168f0a5efc5be191426
%global date    20200930
%global forgeurl https://github.com/lerno/naga

%forgemeta

URL:            %{forgeurl}
Release:        alt1_15jpp11
Summary:        Simplified Java NIO asynchronous sockets

License:        MIT
Source0:        %{forgesource}

[...]

%prep
%forgesetup

=cut

# converted from Lua code used by macros.forge and derivatives

# Computes the suffix of a version string, removing vprefix if it matches
# For example with vprefix 1.2.3: 1.2.3.rc2 → .rc2 but 1.2.30 → 1.2.30 not 0
sub getversionsuffix {
    my ($vstring,$vprefix)=@_;
    my $vprefix_strlen=length($vprefix);
    my $vstring_tail=substr($vstring, - $vprefix_strlen);
    #print STDERR 'ForgeMeta:DEBUG:vstring=',$vstring,' vprefix=',$vprefix,' tail=',$vstring_tail,"\n" if $verbose>8;
    if (substr($vstring, 0, $vprefix_strlen) eq $vprefix and $vstring_tail !~ /^\.?\d/) {
	#print STDERR 'ForgeMeta:DEBUG:return tail=',$vstring_tail,"\n" if $verbose>8;
	return $vstring_tail;
    } else {
	return $vstring;
    }
}

my $urlpatterns = {
    'gitlab' => {
      'pattern'     => qr'https://[^/]+/[^/]+/[^/#?]+',
      'description' => 'https://(…[-.])gitlab[-.]…/owner/repo'},
    'pagure' => {
      'pattern'     => qr'https://[^/]+/[^/#?]+',
      'description' => 'https://pagure.io/repo'},
    'pagure_ns' => {
      'pattern'     => qr'https://[^/]+/[^/]+/[^/#?]+',
      'description' => 'https://pagure.io/namespace/repo'},
    'pagure_fork' => {
      'pattern'     => qr'https://[^/]+/fork/[^/]+/[^/#?]+',
      'description' => 'https://pagure.io/fork/owner/repo'},
    'pagure_ns_fork' => {
      'pattern'     => qr'https://[^/]+/fork/[^/]+/[^/]+/[^/#?]+',
      'description' => 'https://pagure.io/fork/owner/namespace/repo'},
    'gitea.com' => {
      'pattern'     => qr'https://[^/]+/[^/]+/[^/#?]+',
      'description' => 'https://gitea.com/owner/repo'},
    'github' => {
      'pattern'     => qr'https://[^/]+/[^/]+/[^/#?]+',
      'description' => 'https://(…[-.])github[-.]…/owner/repo'},
    'code.googlesource.com' => {
      'pattern'     => qr'https://code.googlesource.com/[^#?]*[^/#?]+',
      'description' => 'https://code.googlesource.com/…/repo'},
    'bitbucket.org' => {
      'pattern'     => qr'https://[^/]+/[^/]+/[^/#?]+',
      'description' => 'https://bitbucket.org/owner/repo'}}
;

sub forge_error {
    return confess;
}

# Check if an identified url is sane
sub checkforgeurl {
    my ($url, $id, $silent)=@_;
    my ($checkedurl,$checkedid);
    my $urlpatterns_id=$urlpatterns->{$id};
    if ($urlpatterns_id) {
	if ($url =~ ($urlpatterns_id->{'pattern'})) {
	    $checkedurl = $&;
	    $checkedid = $id;
	} else {
	    if (not $silent) {
		&forge_error($id , ' URLs must match ' , $urlpatterns_id->{'description'} , ' !');
	    }
	}
    }
    return ($checkedurl, $checkedid);
}

# Check if an url matches a known forge
sub idforge {
    my ($url, $silent)=@_;
    my ($forgeurl, $forge);
    if ($url ne '') {
	if ($url !~ m!^[^:]+://([^/]+)/!) {
	    if (not $silent) {
		forge_error('URLs must include a protocol such as https:// and a path starting with / !');
	    }
	} else {
	    $forge = $1;
	    if ($forge eq 'pagure.io') {
		if ($url =~ m![^:]+://pagure.io/fork/[^/]+/[^/]+/[^/]+!) {
		    $forge = 'pagure_ns_fork';
		} elsif ($url =~ m![^:]+://pagure.io/fork/[^/]+/[^/]+!) {
		    $forge = 'pagure_fork';
		} elsif ($url =~ m![^:]+://pagure.io/[^/]+/[^/]+!) {
		    $forge = 'pagure_ns';
		} elsif ($url =~ m![^:]+://pagure.io/[^/]+!) {
		    $forge = 'pagure';
		}
	    } elsif ($forge =~ m/^gitlab[\.-]/ or $forge =~ m/[\.-]gitlab[\.]/) {
		$forge = 'gitlab';
	    } elsif ($forge =~ m/^github[\.-]/ or $forge =~ m/[\.-]github[\.]/) {
		$forge = 'github';
	    }
	    # TODO:
	    #($forgeurl, $forge) = &checkforgeurl($url, $forge, $silent);
	    ($forgeurl, $forge) = ($url, $forge);
	}
    }
    return ($forgeurl, $forge);
}

sub macro_zalias {
    my ($macros, $listref)=@_;
    foreach my $name (@$listref) {
	my $name0=$name.'0';
	my $has_name=defined ($macros->get_raw($name));
	my $has_name0=defined ($macros->get_raw($name0));
	if ($has_name and not $has_name0) {
	    $macros->set_raw($name0,$macros->get_raw($name));
	} elsif ($has_name0 and not $has_name) {
	    $macros->set_raw($name,$macros->get_raw($name0));
	}
    }
}

sub macro_safeset {
    my ($macros, $name, $value)=@_;
    if (!defined ($macros->get_raw($name))) {
	print STDERR 'ForgeMeta:DEBUG:macro_safeset:set ',$name,'=',$value,"\n" if $verbose>8;
	$macros->set_raw($name,$value);
    }
}

sub macro_explicitset {
    my ($macros, $name, $value)=@_;
    if (!defined($value) or $value eq '') {
	print STDERR 'ForgeMeta:DEBUG:macro_explicitset:deleted ',$name,"\n" if $verbose>8;
	$macros->del_raw($name);
    } else {
	print STDERR 'ForgeMeta:DEBUG:macro_explicitset:set ',$name,'=',$value,"\n" if $verbose>8;
	$macros->set_raw($name,$value);
    }
}

sub __get_match {
    my ($string, $pattern)=@_;
    $string =~ $pattern;
    return $1;
}

# The forgemeta macro main processing function
# See the documentation in the macros.forge file for argument description
# Also called directly by gometa
# expliciturl (-u) is deprecated
sub meta {
    my ($macros, $suffix, $silent, $expliciturl_u_opt) = @_;
    die 'not defined suffix' if not defined($suffix);
    my $ismain = ($suffix eq '' or $suffix eq '0');
    if ($ismain) {
	&macro_zalias(['forgeurl', 'forgesource', 'forgesetupargs',
                      'archivename', 'archiveext', 'archiveurl',
                      'topdir', 'extractdir', 'repo', 'owner', 'namespace',
                      'scm', 'tag', 'commit', 'shortcommit', 'branch', 'version',
                      'date', 'distprefix']);
    }
    print STDERR 'ForgeMeta:DEBUG:enered meta, suffix=',$suffix,' ismain=',($ismain ? '1':'0'),"\n" if $verbose>2;
    # Packaging a moving branch is quite a bad idea, but since at least Gitlab
    # will treat branches and tags the same way better support branches explicitly
    # than have packagers hijack %{tag} to download branch states
    my %spec;
    foreach my $v ('forgeurl','tag','commit','branch','version') {
	$spec{$v} = $macros->macro_subst('%'.$v . $suffix);
    }
    my $forgeurl_expanded=$spec{'forgeurl'};
    print STDERR 'ForgeMeta:DEBUG:forgeurl_expanded=',$forgeurl_expanded,' spec=',Dumper(\%spec),"\n" if $verbose>8;
    my $variables = {
    'default' => {
      'scm'         => 'git',
      'archiveext'  => 'tar.bz2',
      'repo'        => &__get_match($forgeurl_expanded, qr'^[^:]+://[^/]+/[^/]+/([^/?#]+)'),
      'archivename' => '%{repo'         . $suffix . '}-%{ref'           . $suffix . '}',
      'topdir'      => '%{archivename'  . $suffix . '}' },
    'gitlab' => {
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/-/archive/%{ref' . $suffix . '}/%{archivename' . $suffix . '}.%{archiveext' . $suffix . '}' },
    'pagure' => {
      'archiveext'  => 'tar.gz',
      'repo'        => &__get_match($forgeurl_expanded, qr'^[^:]+://[^/]+/([^/?#]+)'),
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/archive/%{ref'   . $suffix . '}/%{archivename' . $suffix . '}.%{archiveext' . $suffix . '}' },
    'pagure_ns' => {
      'archiveext'  => 'tar.gz',
      'namespace'   => &__get_match($forgeurl_expanded, qr'^[^:]+://[^/]+/([^/]+)/[^/?#]+'),
      'repo'        => &__get_match($forgeurl_expanded, qr'^[^:]+://[^/]+/[^/]+/([^/?#]+)'),
      'archivename' => '%{namespace'    . $suffix . '}-%{repo'          . $suffix . '}-%{ref'         . $suffix . '}',
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/archive/%{ref'   . $suffix . '}/%{archivename' . $suffix . '}.%{archiveext' . $suffix . '}' },
    'pagure_fork' => {
      'archiveext'  => 'tar.gz',
      'owner'       => &__get_match($forgeurl_expanded, qr'https://[^/]+/fork/([^/]+)/[^/?#]+'),
      'repo'        => &__get_match($forgeurl_expanded, qr'https://[^/]+/fork/[^/]+/([^/?#]+)'),
      'archivename' => '%{owner'        . $suffix . '}-%{repo'          . $suffix . '}-%{ref'         . $suffix . '}',
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/archive/%{ref'   . $suffix . '}/%{archivename' . $suffix . '}.%{archiveext' . $suffix . '}' },
    'pagure_ns_fork' => {
      'owner'       => &__get_match($forgeurl_expanded, qr'https://[^/]+/fork/([^/]+)/[^/]+/[^/?#]+'),
      'namespace'   => &__get_match($forgeurl_expanded, qr'https://[^/]+/fork/[^/]+/([^/]+)/[^/?#]+'),
      'repo'        => &__get_match($forgeurl_expanded, qr'https://[^/]+/fork/[^/]+/[^/]+/([^/?#]+)'),
      'archivename' => '%{owner'        . $suffix . '}-%{namespace'     . $suffix . '}-%{repo'        . $suffix . '}-%{ref'        . $suffix . '}',
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/archive/%{ref'   . $suffix . '}/%{archivename' . $suffix . '}.%{archiveext' . $suffix . '}' },
    'gitea.com' => {
      'archiveext'  => 'tar.gz',
      'archivename' => '%{fileref'      . $suffix . '}',
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/archive/%{ref'   . $suffix . '}.%{archiveext' . $suffix . '}',
      'topdir'      => '%{repo}' },
    'github' => {
      'archiveext'  => 'tar.gz',
      'archivename' => '%{repo'         . $suffix . '}-%{fileref'       . $suffix . '}',
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/archive/%{ref'   . $suffix . '}/%{archivename' . $suffix . '}.%{archiveext' . $suffix . '}' },
    'code.googlesource.com' => {
      'archiveext'  => 'tar.gz',
      'repo'        => &__get_match($forgeurl_expanded, qr'^[^:]+://.+/([^/?#]+)'),
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/+archive/%{ref'  . $suffix . '}.%{archiveext'  . $suffix . '}',
      'topdir'      => '' },
    'bitbucket.org' => {
      'shortcommit' => substr($macros->macro_subst('%commit'.$suffix), 0, 12),
      'owner'       => &__get_match($forgeurl_expanded, qr'^[^:]+://[^/]+/([^/?#]+)'),
      'archivename' => '%{owner'        . $suffix . '}-%{repo'          . $suffix . '}-%{shortcommit' . $suffix . '}',
      'archiveurl'  => '%{forgeurl'     . $suffix . '}/get/%{ref'       . $suffix . '}.%{archiveext'  . $suffix . '}' } }
    ;

    # Compute the reference of the object to fetch
    my $isrelease = 0;
    my $ref;
    if ($spec{'tag'}) {
	$ref = '%{?tag'     . $suffix . '}';
    } elsif  ($spec{'commit'}) {
	$ref = '%{?commit'  . $suffix . '}';
    } elsif  ($spec{'branch'}) {
	$ref = '%{?branch'  . $suffix . '}';
    } else {
	$ref = '%{?version' . $suffix . '}';
	$isrelease = 1;
    }
    if (not $macros->macro_subst($ref)) {
	if ($suffix eq '') {
	    die 'ForgeMetaError:You need to define Version:, %{commit} or %{tag} before the macro invocation !';
	} else {
	    die 'ForgeMetaError:You need to define %{version', $suffix, '}, %{commit', $suffix, '} or %{tag', $suffix, '} before the macro invocation !';
	}
    }
    &macro_safeset($macros, 'ref' . $suffix, $ref);

    my $forgeurl = $spec{'forgeurl'};
    # For backwards compatibility only
    if ($expliciturl_u_opt) {
      #'%{warn:-u use in %%forgemeta is deprecated, use -z instead to select a separate set of rpm variables!}')
      $forgeurl = $expliciturl_u_opt;
    }
    my $forge;
    ($forgeurl, $forge) = &idforge($forgeurl, $silent);
    if (!$forge) {
	print STDERR 'ForgeMeta:DEBUG:no forge for forgeurl=',$forgeurl,"\n" if $verbose>2;
    } else {
	print STDERR 'ForgeMeta:DEBUG:forge=',$forge,' for forgeurl=',$forgeurl,"\n" if $verbose>2;
	&macro_explicitset($macros, 'forgeurl' . $suffix, $forgeurl);
	# Custom processing of quirky forges that can not be handled with simple variables
	if ($forge eq 'github') {
	    # Workaround the way GitHub injects 'v's before some version strings (but not all!)
	    # To package one of the minority of sane GitHub projects that do not munge their version
	    # strings set tag to %{version} in your spec
	    my $fileref = $ref;
	    my $fileref_expanded=$macros->macro_subst($ref);
	    if ($ref eq '%{?version'  . $suffix . '}') {
		$ref = 'v' . $ref;
	    } elsif  ($fileref ne '%{?commit' . $suffix . '}' and $fileref=~/^v\d/) {
		$fileref = $fileref_expanded=~s/^v//r;
	    } elsif  ($fileref_expanded=~m!/!) {
		$fileref = $fileref_expanded=~s,/,-,gr;
	    }
	    &macro_safeset($macros, 'fileref' . $suffix, $fileref);
	} elsif  ($forge eq 'gitea.com') {
	    # Workaround the way gitea mangles /s in ref names
	    my $fileref = $ref;
	    my $fileref_expanded=$macros->macro_subst($ref);
	    $fileref = $fileref_expanded=~s,/,-,gr;
	    &macro_safeset($macros, 'fileref' . $suffix, $fileref);
	} elsif  ($forge eq 'code.googlesource.com') {
	    if ($ref eq '%{?version'  . $suffix . '}') {
		$ref = 'v' . $ref;
	    }
	} elsif  ($forge eq 'bitbucket.org') {
	    if (!$spec{'commit'}) {
		forge_error('All BitBucket URLs require commit value knowledge: you need to define %{commit}!');
	    }
	}
	&macro_safeset($macros, 'ref' . $suffix, $ref);
	# Mass setting of the remaining variables
	my $_variables_forge=$variables->{$forge};
	foreach my $k (keys(%$_variables_forge)) {
	    my $val = $_variables_forge->{$k};
	    &macro_safeset($macros, $k . $suffix, $val);
	    #$spec{$k}//=$val;
	}
	my $_variables_default=$variables->{'default'};
	foreach my $k (keys(%$_variables_default)) {
	    my $val = $_variables_default->{$k};
	    &macro_safeset($macros, $k . $suffix, $val);
	    #$spec{$k}//=$_val;
	}
    }
    # Generic rules
    foreach my $k ('archiveurl','archivename','archiveext','topdir') {
	my $val=$macros->macro_subst('%{?' . $k . $suffix . '}');
	print STDERR 'ForgeMeta:DEBUG:set spec:k=',$k,' val=',$val,"\n" if $verbose>2;
	$spec{$k}//=$val;
    }
    # Source URL processing (computing the forgesource spec variable)
    my $forgesource = '%{archiveurl' . $suffix . '}';
    $spec{'archiveurl'}=~m!/([^/]+)$!;
    my $archiveurl_end=$1;
    if ($verbose>2) {
	if ($archiveurl_end) {
	    print STDERR 'ForgeMeta:DEBUG:suffix=',$suffix,' archiveurl_end=',$archiveurl_end,"\n";
	} else {
	    print STDERR 'ForgeMeta:DEBUG:suffix=',$suffix,' no archiveurl_end for spec[archiveurl]=',$spec{'archiveurl'},"\n";
	}
    }
    if ($archiveurl_end ne $spec{'archivename'} . '.' . $spec{'archiveext'}) {
	$forgesource = '%{?archiveurl' . $suffix . '}#/%{?archivename' . $suffix . '}.%{archiveext' . $suffix . '}';
    }
    &macro_safeset($macros, 'forgesource' . $suffix, $forgesource);
    # Setup processing      (computing the forgesetup and extractdir variables)
    my $forgesetupargs = '-n %{extractdir' . $suffix . '}';
    my $extractdir     = '%{topdir'        . $suffix . '}';
    if (not $spec{'topdir'}) {
	$forgesetupargs = '-c ' . $forgesetupargs;
	$extractdir = '%{archivename' . $suffix . '}';
    }
    if (not $ismain) {
	if ($spec{'topdir'}) {
	    $forgesetupargs = '-T -D -b ' . $suffix . ' ' . $forgesetupargs;
	} else {
	    $forgesetupargs = '-T -D -a ' . $suffix . ' ' . $forgesetupargs;
	}
    }
    &macro_safeset($macros, 'forgesetupargs' . $suffix, $forgesetupargs);
    &macro_safeset($macros, 'extractdir'     . $suffix, $extractdir);
    # dist processing       (computing the correct prefix for snapshots)
    my $distprefix = '';
    if (not $isrelease) {
	$distprefix=lc($macros->macro_subst($ref));
    }
    if ($ref eq '%{?commit' . $suffix . '}') {
	$distprefix = substr($distprefix, 0, 7);
    } elsif ($ref ne '%{?branch' . $suffix . '}') {
	#$distprefix = string.gsub(distprefix,      "[%p%s]+", '.')
	$distprefix=~s/[^\w\d]+/./g;
	#distprefix = string.gsub(distprefix, '^' .. string.lower(rpm.expand('%{?repo}')) .. "%.?", '')
	my $lcrepo=lc($macros->macro_subst('%{?repo}'));
	$distprefix=~s/^${lcrepo}\.?//;
	my $version_expanded=$macros->macro_subst('%{version}');
	my $v = $version_expanded=~s/[^\w\d]+/./gr;
	for my $p ('','v','v.','version','version.','tags.v', 'tags.v.') {
	    $distprefix = &getversionsuffix($distprefix, $p . $v);
	}
	$distprefix=~s/^\.//;
    }

    if ($distprefix and $distprefix ne '') {
	$distprefix = '%{scm'     . $suffix . '}' . $distprefix;
	my $date = $macros->macro_subst('%{?date' . $suffix . '}');
	if ($date) {
	    $distprefix = $date . $distprefix;
	} else {
	    my $src_archive=$macros->macro_subst('%{_sourcedir}/%{archivename'.$suffix.'}.%{archiveext'.$suffix.'}');
	    if (-r $src_archive) {
		$distprefix = `date '+%Y%m%d' -u -r $src_archive`. $distprefix;
	    }
	    $distprefix = '.' . $distprefix;
	}
    }
    my $version_expanded=$macros->macro_subst('%{?version}');
    if ($spec{'version'} and $spec{'version'} ne '0' and
	$spec{'version'} ne $version_expanded) {
	$distprefix = '.%{version' . $suffix . '}' . $distprefix;
    }
    if ($macros->macro_subst($distprefix) ne '') {
	if (not $ismain) {
	    $distprefix =~ s/^\./.s/;
	}
	&macro_safeset ($macros,'distprefix' . $suffix, $distprefix);
    }
    if ($ismain) {
	&macro_zalias($macros,['forgeurl', 'forgesource', 'forgesetupargs',
                      'archivename', 'archiveext', 'archiveurl',
                      'topdir', 'extractdir', 'repo', 'owner', 'namespace',
			    'scm', 'shortcommit', 'distprefix']);
    }
    return;
}


=head

    # Final spec variable summary if the macro was called with -i
    if ($informative) {
    rpm.expand("%{echo:Packaging variables read or set by %%forgemeta}")
    fedora.echovars({"forgeurl", "forgesource", "forgesetupargs",
                        "archivename", "archiveext", "archiveurl",
                        "topdir", "extractdir", "repo", "owner", "namespace",
                        "scm", "tag", "commit", "shortcommit", "branch", "version",
                        "date", "distprefix"}, suffix)
    fedora.echovars({"dist"},"")
    rpm.expand("%{echo:  (snapshot date is either manually supplied or computed once %%{_sourcedir}/%%{archivename" . $suffix . "}.%%{archiveext" . $suffix . "} is available)}")
  }

=cut

# TODO: defines better be generic (e.q. use %{version}, not expanded value, etc)
# TODO: no need to create defines if they are already present in spec
sub forgemeta_defines {
    my ($macros,$suffix)=@_;
    my @outlist;
    foreach my $key ('forgeurl', 'forgesource', #'forgesetupargs',
		     'archivename', 'archiveext', 'archiveurl',
		     'topdir', 'extractdir', 'repo', 'owner', 'namespace',
		     'scm', 'tag', 'commit', 'shortcommit', 'branch', 'version',
		     'date', 'distprefix') {
	my $val_expanded=$macros->macro_subst('%'.$key);
	if (defined $val_expanded and $val_expanded ne '') {
	    push @outlist, '%global '.$key.$suffix.' '.$val_expanded."\n";
	} else {
	    push @outlist, '#global '.$key.$suffix.' '.'%nil'."\n";
	}
    }
    return @outlist;
}

1;

__END__
